import pytest

from ruamel.yaml.compat import ordereddict

import rasax.community.utils
from rasax.community import config
from rasax.community.services.settings_service import (
    SettingsService,
    default_stack_config,
)

TEST_PROJECT = "test project"
TEST_TEAM = "test team"


def default_environments_config():
    stack_config = ordereddict(
        [
            (
                "production",
                ordereddict(
                    [
                        ("url", "http://rasa-production:5005"),
                        ("token", config.rasa_token),
                    ]
                ),
            ),
            (
                "development",
                ordereddict(
                    [
                        ("url", "http://rasa-development:5005"),
                        ("token", config.rasa_token),
                    ]
                ),
            ),
            (
                "worker",
                ordereddict(
                    [("url", "http://rasa-worker:5005"), ("token", config.rasa_token)]
                ),
            ),
        ]
    )

    return {"environments": ordereddict([("rasa", stack_config)])}


def test_project_init(settings_service: SettingsService):
    settings_service.init_project(TEST_TEAM, TEST_PROJECT)

    actual = settings_service.get(TEST_TEAM, TEST_PROJECT)

    assert actual["config"] == default_stack_config()


def test_save_config(settings_service: SettingsService):
    test_config = default_stack_config()
    test_config["extra"] = "extra value"

    settings_service.save_config(TEST_TEAM, TEST_PROJECT, test_config)

    actual = settings_service.get(TEST_TEAM, TEST_PROJECT)["config"]

    assert actual == test_config


def test_get_stack_services(settings_service: SettingsService):
    default_env = default_environments_config()["environments"]["rasa"]

    settings_service.save_environments_config(
        TEST_PROJECT, default_environments_config().get("environments")
    )

    actual = settings_service.stack_services(TEST_PROJECT)

    assert len(actual) == len(default_env)

    for k, v in default_env.items():
        assert actual[k].async_stack_endpoint.token == v["token"]
        assert actual[k].async_stack_endpoint.url == v["url"]


def test_inspect_valid_environment_config():
    default_env = default_environments_config()["environments"]

    assert SettingsService.inspect_environments_config(default_env) is None


def test_invalid_environment_config_without_rasa_key():
    default_env = default_environments_config()["environments"]
    default_env["other_key"] = default_env.pop("rasa")

    with pytest.raises(ValueError):
        SettingsService.inspect_environments_config(default_env)


@pytest.mark.parametrize("environment", ["production", "worker"])
def test_invalid_environment_config_without_minimal_instances(environment):
    default_env = default_environments_config()["environments"]
    del default_env["rasa"][environment]

    with pytest.raises(ValueError):
        SettingsService.inspect_environments_config(default_env)


def test_extra_environments_with_enterprise():
    default_env = default_environments_config()["environments"]
    default_env["rasa"]["another"] = default_env["rasa"]["production"]

    assert SettingsService.inspect_environments_config(default_env) is None


def test_extra_environments_with_rasa_ce():
    default_env = default_environments_config()["environments"]
    default_env["rasa"]["another"] = default_env["rasa"]["production"]

    # path to pretend we don't have enterprise installed
    rasax.community.utils.is_enterprise_installed = lambda: False

    with pytest.raises(ValueError):
        SettingsService.inspect_environments_config(default_env)

    # revert patch
    rasax.community.utils.is_enterprise_installed = lambda: True


def test_save_environment_config(settings_service: SettingsService):
    default_env = default_environments_config()["environments"]
    test_project = "default"

    settings_service.save_environments_config(test_project, default_env)

    assert settings_service.get_environments_config(test_project)


def test_delete_environment_config(settings_service: SettingsService):
    default_env = default_environments_config()["environments"]
    test_project = "default"

    settings_service.save_environments_config(test_project, default_env)

    settings_service.delete_environment_config("production", test_project)

    actual = settings_service.get_environments_config(test_project)

    assert "production" not in actual["environments"]["rasa"]
